--  file    : rat.e
--  author  : jiri babor
--  email   : jbabor@paradise.net.nz
--  project : mouse routines for graphics modes
--  tool    : euphoria 2.1
--  date    : 99-12-06
--  version : 2.20


include machine.e

global constant
    LEFT_BUTTON_PRESSED=1,
    RIGHT_BUTTON_PRESSED=2,
    MIDDLE_BUTTON_PRESSED=4

global integer
    rat_button,
    rat_x,
    rat_y,
    rat_visible     -- boolean, rat pointer visible or hidden (1 or 0)

constant REGS=repeat(0,10)

sequence pointer    -- rle pointer image
integer maxx,
    minx,
    maxy,
    miny,
    hot_x,
    hot_y,
    last_x,         -- last drawn horizontal coordinate
    last_y          -- last drawn vertical coordinate


global function compile_rat_pointer(
    integer x,      -- 'hot' x
    integer y,      -- 'hot' y
    sequence s      -- bitmap, 2-d color pointer image
    )
    -- run length encode pointer image for faster display later

    sequence si,u,ui
    integer h,inrun,count,w,k

    h=length(s)
    u={}
    for i=1 to length(s) do
	inrun=0
	count=0
	si=s[i]
	ui={}
	w=length(s[i])
	for j=1 to w do
	    if si[j]=0 then
		if inrun then
		    ui &= count & si[k..j-1]
		    count=1
		    inrun=0
		else
		    count+=1
		end if
	    else
		if inrun then
		    count+=1
		else
		    ui &= count
		    k=j
		    count=1
		    inrun=1
		end if
	    end if
	end for
	if inrun then
	    ui &= count & si[k..w]
	end if
	u=append(u,ui)
    end for
    return prepend(u, {x,y})
end function -- compile_rat_pointer

procedure draw_rat()
    -- display rle pointer image

    sequence rlei
    integer j,run,n,u,x,y

    x=last_x-hot_x
    y=last_y-hot_y
    for i=1 to length(pointer) do
	rlei=pointer[i]
	j=1
	u=x
	run=0
	while j<length(rlei) do
	    n=rlei[j]
	    if run then
		pixel(xor_bits(get_pixel({u,y,n}),rlei[j+1..j+n]),{u,y})
		j+=n+1
	    else
		j+=1
	    end if
	    run=not run
	    u+=n
	end while
	y+=1
    end for
end procedure -- draw_rat

global procedure show_rat()
    if not rat_visible then draw_rat() end if
    rat_visible=1
end procedure -- show_rat

global procedure hide_rat()
    if rat_visible then draw_rat() end if     -- erase
    rat_visible=0
end procedure -- hide_rat

global procedure rat()
    sequence regs
    integer m

    regs=REGS
    regs[REG_AX]=#B
    regs=dos_interrupt(#33,regs)    -- relative motion counters

    m=regs[REG_CX]
    if m then
	rat_x += m - 2 * and_bits(m, #8000)
	if rat_x<minx then rat_x=minx
	elsif rat_x>maxx then rat_x=maxx
	end if
    end if

    m=regs[REG_DX]
    if m then
	rat_y += m - 2 * and_bits(m, #8000)
	if rat_y<miny then rat_y=miny
	elsif rat_y>maxy then rat_y=maxy
	end if
    end if

    if rat_visible and (last_x!=rat_x or last_y!=rat_y) then
	draw_rat()                  -- erase previous image
	last_x=rat_x
	last_y=rat_y
	draw_rat()                  -- show again
    else
	last_x=rat_x
	last_y=rat_y
    end if

    regs[REG_AX]=3
    regs=dos_interrupt(#33,regs)    -- poll rat
    rat_button=regs[REG_BX]
end procedure -- rat

global procedure confine_rat(
    sequence p1,    -- left top corner of screen cage
    sequence p2     -- right bottom corner
    )
    minx=p1[1]
    maxx=p2[1]
    miny=p1[2]
    maxy=p2[2]
    if rat_x<minx then
	rat_x=minx
    elsif rat_x>maxx then
	rat_x=maxx
    end if
    if rat_y<miny then
	rat_y=miny
    elsif rat_y>maxy then
	rat_y=maxy
    end if
end procedure -- confine_rat

global procedure set_rat_xy(integer x, integer y)
    -- places the rat pointer where you want it

    if x<minx then
	x=minx
    elsif x>maxx then
	x=maxx
    end if
    if y<miny then
	y=miny
    elsif y>maxy then
	y=maxy
    end if
    rat_x=x
    rat_y=y
    rat()
end procedure  -- set_rat_xy

global function get_pointer()
    return  {{hot_x,hot_y}}&pointer
end function
global procedure rat_pointer(
    sequence p  -- rle pointer image with prepended hot spot coordinates
    )
    -- select new pointer

    integer flag

    flag=rat_visible    -- save current rat visibility
    hide_rat()          -- make rat invisible
    hot_x=p[1][1]-1
    hot_y=p[1][2]-1
    pointer=p[2..length(p)]
    if flag then show_rat() end if  -- make new pointer visible
end procedure -- rat_pointer

global procedure init_rat()
    -- initialize - must be called *after* graphics mode is set
    -- set pointer in middle of screen, hidden
    -- confine rat to visible screen area

    sequence vc

    vc=machine_func(13,0)           -- video_config()
    rat_button=0
    rat_x=floor(vc[5]/2)
    last_x=rat_x
    rat_y=floor(vc[6]/2)
    last_y=rat_y
    minx=0
    maxx=vc[5]-1
    miny=0
    maxy=vc[6]-1
end procedure -- init_rat

global procedure wait_while (integer pressed_button)
    -- wait while specified rat button remains pressed

    while and_bits(rat_button, pressed_button) do
	rat()      -- keep rat pointer alive
    end while
end procedure -- wait_while

global function rat_zone(sequence zones)
    -- scan zones for rat pointer presence
    -- zones is sequence of rectangular zones to be tracked
    -- each zone is defined by {x1y1, x2y2) sequence, where x1y1 and
    -- x2y2 are top_left & bottom_right diagonal points respectively
    -- return index of the *first* zone that includes rat pointer, or
    -- zero if none found

    sequence zi

    for i=1 to length(zones) do
	zi=zones[i]
	if rat_x>=zi[1][1] and
	   rat_x<=zi[2][1] and
	   rat_y>=zi[1][2] and
	   rat_y<=zi[2][2] then
	   return i
	end if
    end for
    return 0
end function -- rat_zone


-- initialize default rat pointer ------------------------------------

global constant RAT_ARROW = compile_rat_pointer(1, 1,{
	"/",
	"//",
	"/'/",
	"/''/",
	"/'''/",
	"/''''/",
	"/'''''/",
	"/''''''/",
	"/'''''''/",
	"/'''''////",
	"/'''''/",
	"/'//''/",
	"//  /''/",
	"/   /''/",
	"     /''/",
	"     ////"}-32)
rat_visible=0       -- let's make rat initially hidden
rat_pointer(RAT_ARROW)
rat_x=0 rat_y=0 rat_button=0
