
 ---------------------------------------------------------------------
 --                                                                 --
 --                        Topaz v. 1.0                             --
 --                                                                 --
 --            Imbedded Stack-Based Scripting Engine                --
 --                    (a "virtual machine")                        --
 --                                                                 --
 --                                                                 --
 --             Created July 1, 1999 by Rod Jackson.                --
 --                     All rights reserved.                        --
 --                                                                 --
 --                                                                 --
 -- This source code is made available free of charge for all       --
 -- public, private, and commercial use. It has been designed to    --
 -- run under version 2.0 of the Euphoria interpreter by Rapid      --
 -- Development Software. Only authorized modifications or upgrades --
 -- should be distributed unless said modifications or upgrades are --
 -- clearly identified as "non-official" or "unofficial".           --
 --                                                                 --
 --                                                                 --
 -- For more information on this library, please refer to the       --
 -- documentation files, or contact the author at:                   --
 --                                                                 --
 --    rodjackson@bigfoot.com                                       --
 --                                                                 --
 ---------------------------------------------------------------------


 ---------------------
 -- local constants --
 ---------------------

include rat.e
include image.e
include file.e
include get.e
include dos.e
constant MAX_EU_INT = 1073741823 -- the max. legal size of an integer
constant TOPAZ_ABORT_CODE = 'T'  -- this is 84 decimal, 54 hex
global constant -- Thread flags
		FLAGS_KILL1    = 1, -- It's not recommended to kill this thread
		FLAGS_KILL2    = 2, -- You CANNOT kill this thread
		FLAGS_READONLY = 4, -- The flags are read-only
		FLAGS_DONTKEEP = 8, -- Don't keep screen position
		MKTIL_CONVERT  = 64 -- Converts f.p. or bytes < 0 or > 255 to
				    -- another format
global constant MIN_PRIORITY=1
global constant MAX_PRIORITY=32
global constant DEFAULT_PRIORITY=10
-- It is recommended that you always set this to 1, but you might set it to
-- 0 when debugging
global constant errorChecking = 1,   -- Enable the EAD
		enableOtherStuff = 1 -- Enable other EAD functions

 -----------------
 -- local types --
 -----------------

 -- a legal virtual dataspace address
type vaddress (object x)
   if (integer (x)) then
      if (x > 0) then
	 return 1 -- TRUE
      end if
   end if
   return 0 -- FALSE
end type

 ---------------------
 -- local variables --
 ---------------------
global sequence
	 InstructionSpaceTable,   -- the instruction space per thread
	 StackSizeTable,          -- the size of each thread's stack
	 StackTable,              -- the stack of each thread
	 DataSpaceTable,          -- the data space of each thread
	 pointers,
	 executed,
	 ThreadIDTable,           -- a unique ID per thread
	 sIDs,
	 keyIDs,
	 opened,
	 files,
	 vars,
	 vlist,
	 msgs,
	 ipStack,
	 modes,
	 fpos,
	 sIDs2,
	 tmrIDs,
	 timers,
	 positions,
	 keys,
	 SuspendedBy,
	 ThreadFlags,
	 funcs

sequence InstructionPointerTable, -- next instr. to exec. per thread
	 InstructionRoutineTable, -- routine ID per instruction
	 IsSuspendedTable,        -- if each thread is suspended
	 PriorityTable,           -- priority for each thread
	 TimeTable,               -- time for each thread
	 ThreadNames,
	 backtrack

global integer
	 ThreadPointer,       -- indicates the current thread
	 execute,
	 success,
	 key,
	 checkThreadID,
	 stuffID
	 
checkThreadID=-1
stuffID=-1

integer  MaxInstruction,      -- the highest value of all instructions
	 TP_IsTrapped,        -- if the thread pointer is trapped
	 TP_TrapsAllowed,     -- if thread pointer trapping is allowed
	 NumThreads,          -- the number of existing threads
	 NextThreadID,        -- the value to assign a new thread's ID
	 CurrentIP,           -- the IP of the current thread
	 IPWasChanged,        -- if the curr. instr. for the thread
			      --   altered the instruction pointer
	 CurrThreadIsFinished,-- if the current thread has ended
	 TickCount,           -- Current "time"
	 killID,
	 killed,
	 incrementMode

 funcs={}
 backtrack={}
 incrementMode=1
 ThreadFlags={}
 keys={}
 key=-1
 execute=1
 positions={}
 executed={}
 pointers={}
 success=0
 killed=0
global function topazGet_key()
    return key
end function
global function getThreadID(integer i)
    return ThreadIDTable[i]
end function
global function getThreadNames()
    return {ThreadNames,ThreadIDTable}
end function
global function getThreadName(integer id)
    integer i
    if id=-1 then
	return "Main program"
    end if
    i=find(id,ThreadIDTable)
    if i then
	return ThreadNames[i]
    end if
    return ""
end function
global procedure setThreadName(integer id,sequence name)
    integer len
    len=0
    if find(id,ThreadIDTable) then
	for z = 1 to length(name) do
	    if not find(name[z]," \t\r\n"&0) then
		len += 1
	    end if
	end for
	if len then
	    ThreadNames[find(id,ThreadIDTable)]=name
	end if
    end if
end procedure
global function threadID(integer id)
    if id >= 1 and id <= length(ThreadIDTable) then
	return ThreadIDTable[id]
    end if
    return -1
end function
 global procedure showIP()
    puts(1,"Thread IP's:\n")
    for z = 1 to length(pointers) do
	printf(1,"Thread %d: %d\n",{threadID(z),InstructionPointerTable[z]})
    end for
 end procedure
procedure showStack()
    puts(1,'\n')
    for a = 1 to length(StackTable) do
	printf(1,"Stack table belonging to thread %d(%s)\n%d elements\n",
		 {threadID(a),getThreadName(threadID(a)),StackSizeTable[a]})
	for b = 1 to length(StackTable[a]) do
	    if sequence(StackTable[a][b]) then
		print(1,StackTable[a][b])
	    else
		printf(1,"%g",StackTable[a][b])
	    end if
	    if b < length(StackTable[a]) then
		puts(1,' ')
	    end if
	end for
	puts(1,'\n')
    end for
end procedure
 --------------------
 -- local routines --
 --------------------

 procedure Kill_thread(integer id)
    call_proc(killID,{id})
 end procedure
global procedure terminate(integer a)
    integer k
    k=-1
    puts(1,"Press Enter to delete the thread, Escape to delete all"&
	   " threads, or Delete to\nterminate and get an error log\n")
    while k != 13 or k != 27 do
	k=wait_key()
	if k=13 then
	    Kill_thread(a-1)
	    killed=1
	    exit
	elsif k=27 then
	    for z = 2 to NumThreads do
		Kill_thread(z-1)
	    end for
	    killed=1
	    puts(1,"All threads deleted\n")
	    exit
	elsif k='S' or k='s' then
	    showStack()
	elsif k=339 then
	    ? 1/0
	end if
    end while
end procedure
 --
 -- Remove: Returns the given sequence, with the specified element
 --    removed.
 --
function Remove (integer pos, sequence s)

   integer L

   L = length (s)

   if L=0 then
    return s
   end if
   if (pos = 1) then
      return s[2..L]
   elsif (pos = L) then
      return s[1..L-1]
   elsif pos < 1 or pos > L then
    return s
   else
      return s[1..pos-1] & s[pos+1..L]
   end if

end function -- end Remove()


 --
 -- IncrementThreadID: Increments the value of NextThreadID so that
 --    any new threads will be given a unique ID.
 --
procedure IncrementThreadID ()

   integer AttemptedNext

   if (NextThreadID = MAX_EU_INT) then
      AttemptedNext = 0
   else
      AttemptedNext = NextThreadID + 1
   end if

    -- make sure the ID isn't already taken (highly unlikely, but...)
   while (find (AttemptedNext, ThreadIDTable)) do
      if (AttemptedNext = MAX_EU_INT) then
	 AttemptedNext = 0
      else
	 AttemptedNext = AttemptedNext + 1
      end if
   end while

   NextThreadID = AttemptedNext
end procedure -- end IncrementThreadID()


 --
 -- IncrementThreadPointer: Increments the thread pointer so that the
 --    next thread will be executed. Cycles back to the first after
 --    running through all existing threads.
 --
procedure check_thread(integer id,integer ip)
    if checkThreadID > -1 then
	call_proc(checkThreadID,{id,ip})
    end if
end procedure
procedure other_stuff()
    if stuffID > -1 then
	call_proc(stuffID,{})
    end if
end procedure
procedure IncrementThreadPointer ()
   sequence adjustedTime,temp
   integer best
   adjustedTime=TimeTable/PriorityTable
   if incrementMode=1 then
    best=1
    for i=1 to NumThreads do
       if IsSuspendedTable[i]=1 then
	  adjustedTime[i]*=-1
       end if
       if adjustedTime[i]<adjustedTime[best] and adjustedTime[i]>=0 then
	  best=i
       end if
    end for
    backtrack &= best
    if length(backtrack)=200 then
	temp=backtrack backtrack={}
	if equal(temp,repeat(best,200)) then
	    incrementMode=2
	end if
    end if
   elsif incrementMode=2 then
    best=ThreadPointer+1
    if best > NumThreads then
	best=1
    end if
   end if
   if IsSuspendedTable[best]=1 then
    for a = 1 to NumThreads do
	if IsSuspendedTable[a]=0 then
	    best=a exit
	end if
    end for
   end if
   ThreadPointer = best

end procedure -- end IncrementThreadPointer()


 --
 -- InitTopaz: Initializes the internal variables and tables of the
 --    Topaz engine.
 --
procedure InitTopaz ()

    -- initialize the thread-local data tables
   InstructionSpaceTable   = {}
   DataSpaceTable          = {}
   StackTable              = {}
   StackSizeTable          = {}
   ThreadIDTable           = {}
   InstructionPointerTable = {}
   IsSuspendedTable        = {}
   SuspendedBy             = {}
   PriorityTable           = {}
   TimeTable               = {}
   ThreadNames             = {}
    -- initialize the table to instruction code
   InstructionRoutineTable = {}

    -- initialize the internal Topaz variables
   MaxInstruction       = 0
   ThreadPointer        = 1
   TP_IsTrapped         = 0 -- FALSE
   TP_TrapsAllowed      = 1 -- TRUE
   NumThreads           = 0
   NextThreadID         = 0
   CurrentIP            = 0
   IPWasChanged         = 0 -- FALSE
   CurrThreadIsFinished = 0 -- FALSE
   TickCount            = 0

end procedure -- end InitTopaz()


 --
 -- TopazError: Displays a Topaz-specific error message in a
 --    standardized format, then aborts with a Topaz-specific error
 --    code.
 --
procedure TopazError (object  LorPorF,     -- if func., proc. or lib.
		      sequence RoutineName, -- name of routine error
					    --   was found in
		      sequence ErrMessage)  -- the specific error
					    --   message

   sequence EntireMessage
    if sequence(LorPorF) then
	LorPorF=LorPorF[1]
    end if
   EntireMessage = "\nTopaz error in "

   if (LorPorF = 'l') or (LorPorF = 'L') then
      EntireMessage = EntireMessage & "library -\n   " &
		      ErrMessage & "\n"
   else
      if (LorPorF = 'p') or (LorPorF = 'P') then
	 EntireMessage = EntireMessage & "procedure"
      else
	 EntireMessage = EntireMessage & "function"
      end if

      EntireMessage = EntireMessage & ": " & RoutineName & " -\n   " &
		      ErrMessage & "\n"
   end if

   puts (1, EntireMessage)

   abort (TOPAZ_ABORT_CODE)

end procedure -- end TopazError


 --
 -- TopazExec: Executes the given instruction by calling the
 --    appropriate routine and passing all parameters.
 --
global procedure TopazExec (sequence FullInstr)

   object TheInstr

   TheInstr = FullInstr[1]

   if (not (integer (TheInstr))) then
      TopazError ('p', "TopazExec", "Unknown instruction encountered")
   end if

   if execute then
   call_proc (InstructionRoutineTable[TheInstr],
	      {FullInstr[2..length(FullInstr)]})
   end if
   execute=1

end procedure -- end TopazExec

 ------------------------------
 -- primary library routines --
 ------------------------------

 --
 -- add_instruction: Adds a new instruction to the Topaz engine by
 --    registering the routine ID of the code that will be executed
 --    when the instruction is encountered. After being added, the
 --    instruction can immediately be utilized.
 --
global function add_instruction (integer RoutineID)

   MaxInstruction = MaxInstruction + 1
   InstructionRoutineTable = InstructionRoutineTable & RoutineID
   return MaxInstruction

end function -- end add_instruction()


 --
 -- allow_tp_traps: Tells the Topaz engine to honor all subsequent
 --    instructions to trap the thread pointer.
 --
global procedure allow_tp_traps ()

   TP_TrapsAllowed = 1 -- TRUE

end procedure -- end allow_tp_traps()


 --
 -- disallow_tp_traps: Tells the Topaz engine to untrap the thread
 --    pointer if trapped, and to ignore all subsequent instructions
 --    to trap it.
 --
global procedure disallow_tp_traps ()

   TP_TrapsAllowed = 0 -- FALSE
   TP_IsTrapped    = 0 -- FALSE

end procedure -- end disallow_tp_traps()


 --
 -- is_tp_trapped: Returns 1 (TRUE) if the thread pointer is trapped,
 --    otherwise returning 0 (FALSE).
 --
global function is_tp_trapped ()

   return TP_IsTrapped

end function -- end is_tp_trapped()


 --
 -- is_stack_empty: Returns 1 (TRUE) if the current thread's stack is
 --    empty, otherwise returning 0 (FALSE).
 --
global function is_stack_empty ()

   return (compare (StackTable[ThreadPointer], {}) = 0)

end function -- end is_stack_empty()


 --
 -- get_vaddress: Returns the object at the specified virtual address
 --    of the data space of the current thread.
 --
global function get_vaddress (vaddress Addr)

   return DataSpaceTable[ThreadPointer][Addr]

end function -- end get_vaddress()


 --
 -- put_vaddress: Puts the given object into the data space of the
 --    current thread, at the specified virtual address.
 --
global procedure put_vaddress (object Obj, vaddress Addr)

   DataSpaceTable[ThreadPointer][Addr] = Obj

end procedure -- end put_vaddress()


 --
 -- pop_stack: Pops the top element of the stack of the current
 --    thread, shrinking the stack. The popped element is returned.
 --
global function pop_stack ()

   object   temp
   integer  stacksize

   stacksize = StackSizeTable[ThreadPointer]
   
   if stacksize=0 then
    return 0
   end if

   temp = StackTable[ThreadPointer][stacksize]

   stacksize = stacksize - 1

   StackTable[ThreadPointer] =
	 StackTable[ThreadPointer][1..stacksize]

   StackSizeTable[ThreadPointer] = stacksize

   return temp

end function -- end pop_stack()


 --
 -- push_stack: Pushes the given object onto the stack of the current
 --    thread.
 --
global procedure push_stack (object Obj)

   StackTable[ThreadPointer] = append (StackTable[ThreadPointer], Obj)

   StackSizeTable[ThreadPointer] = StackSizeTable[ThreadPointer] + 1

end procedure -- end push_stack()


 --
 -- new_thread: Creates a new thread, giving it its own data space,
 --    instruction space, and stack. The code to be placed into the
 --    instruction space must be provided, as must the size of the
 --    data space (in number of elements).
 --
global function new_thread (integer  DataSpaceSize,integer priority,
			    sequence InstructionCode)

   sequence vc
   integer ThreadID,totalPriority,timeAdjust,gfxMode

   vc=video_config()
   gfxMode=vc[VC_MODE]
   ThreadID = NextThreadID
   ThreadIDTable = ThreadIDTable & ThreadID
   IncrementThreadID ()

   InstructionSpaceTable = append
			     (InstructionSpaceTable, InstructionCode)
   DataSpaceTable = append (DataSpaceTable, repeat (0, DataSpaceSize))
   positions = append (positions, get_position()&gfxMode)
   InstructionPointerTable = InstructionPointerTable & 1
   pointers &= 1
   StackTable = append (StackTable, {})
   executed &= 0
   ThreadFlags &= 0
   StackSizeTable = StackSizeTable & 0
   IsSuspendedTable = IsSuspendedTable & 0 -- FALSE
   SuspendedBy &= -1
   totalPriority=0
   for i=1 to length(PriorityTable) do
      if IsSuspendedTable[i]=0 then totalPriority+=PriorityTable[i] end if
   end for
   if totalPriority=0 then
      timeAdjust=0
   else
      timeAdjust=floor(priority/totalPriority*TickCount+.5)
   end if
   if priority<MIN_PRIORITY then priority=MIN_PRIORITY end if
--    elsif priority>MAX_PRIORITY then priority=MAX_PRIORITY end if
   PriorityTable = PriorityTable & priority
   TimeTable = TimeTable & timeAdjust
   TickCount += timeAdjust
   NumThreads = NumThreads + 1

   ThreadNames &= {sprintf("Thread #%d",ThreadID)}
   return ThreadID

end function -- end new_thread()

 --
 -- kill_thread: kills the specified thread, removing its stack,
 --    instruction space and data space.
 --
global function thread_exists(integer id)
    return find(id,ThreadIDTable) != 0
end function
global procedure kill_thread (integer ThreadID)

   integer thread_pos

    -- find out at what position the specified thread is found
   thread_pos = find (ThreadID, ThreadIDTable)

    -- make sure the thread ID given is legal
   if (thread_pos = 0) then
      --TopazError ("f", "kill_thread", "Unknown thread ID given")
      return
   end if

    -- remove the local data for the thread from the tables
   InstructionSpaceTable = Remove (thread_pos, InstructionSpaceTable)
   InstructionPointerTable = Remove (thread_pos,
				     InstructionPointerTable)
   StackTable = Remove (thread_pos, StackTable)
   StackSizeTable = Remove (thread_pos, StackSizeTable)
   DataSpaceTable = Remove (thread_pos, DataSpaceTable)
   ThreadIDTable = Remove (thread_pos, ThreadIDTable)
   IsSuspendedTable = Remove (thread_pos, IsSuspendedTable)
   SuspendedBy = Remove(thread_pos, SuspendedBy)
   PriorityTable = Remove (thread_pos, PriorityTable)
   TickCount = TickCount - TimeTable[thread_pos]
   TimeTable = Remove (thread_pos, TimeTable )
   ThreadNames = Remove (thread_pos, ThreadNames )
   pointers = Remove(thread_pos,pointers)
   positions = Remove(thread_pos,positions)
   keys=Remove(thread_pos,keys)
   ThreadFlags=Remove(thread_pos,ThreadFlags)
    -- adjust & untrap the thread pointer if necessary
   if (thread_pos < ThreadPointer) then
      ThreadPointer = ThreadPointer - 1
   elsif (thread_pos = ThreadPointer) then
      TP_IsTrapped = 0 -- FALSE
   end if

    -- note that the number of threads is reduced
   NumThreads = NumThreads - 1

    -- make sure we're not past the threads
   if (ThreadPointer > NumThreads) then
      ThreadPointer = 1
   end if

   CurrThreadIsFinished = 0 -- FALSE

end procedure -- end kill_thread()


 --
 -- get_thread_id: Returns the unique ID of the current thread.
 --
global function get_thread_id ()

   return ThreadIDTable[ThreadPointer]

end function -- end get_thread_id()


 --
 -- get_num_threads: Returns the total number of existant threads.
 --
global function get_num_threads ()

   return NumThreads

end function -- end get_num_threads()


 --
 -- inc_ip: Increments the current instruction pointer by the
 --    specified amount, and notes that the instruction pointer was
 --    modified.
 --
global procedure inc_ip (integer increment)

   CurrentIP = CurrentIP + increment

   IPWasChanged = 1 -- TRUE

end procedure -- end inc_ip()
global function get_ip()
    return CurrentIP
end function
global procedure reInit(integer t)
    InstructionPointerTable[t]=1
    StackSizeTable[t]=0
    StackTable[t]={}
    DataSpaceTable[t]=repeat(0,length(DataSpaceTable))
    pointers[t]=1
    IsSuspendedTable[t]=0
    SuspendedBy[t]=-1
    --executed[t]=0
end procedure
global function getIP(integer thread)
    return InstructionPointerTable[thread]
end function
global function getFlags(integer id)
    if find(id,ThreadIDTable) then
	return ThreadFlags[find(id,ThreadIDTable)]
    end if
    return 0
end function
global procedure setFlags(integer id,integer flgs)
    integer i
    i=find(id,ThreadIDTable)
    if i then
	-- Only if flags are not read-only
	if and_bits(ThreadFlags[i],FLAGS_READONLY) != FLAGS_READONLY then
	    ThreadFlags[i]=flgs
	end if
    end if
end procedure
 --
 -- topaz_step: Executes the next instruction of the current thread,
 --    then increments the thread pointer so that next time the next
 --    thread will be executed. If the thread pointer is trapped, it
 --    will not be incremented.
 --
global procedure topaz_step ()

   sequence vc
   integer gfxMode
   object FullInstr

   vc=video_config()
   gfxMode=vc[VC_MODE]
   if enableOtherStuff then
    other_stuff()
   end if
    -- only execute if thread pointer is valid (if not, most likely
    -- there are no existing threads, which we'll handle), and...
   if (ThreadPointer <= NumThreads) then

       -- ...if the thread is not suspended
      if (IsSuspendedTable[ThreadPointer] = 0) then

	 CurrentIP = InstructionPointerTable[ThreadPointer]

	 if (CurrentIP < 0) then
	     -- this means the last command was an 'exec', so we'll
	     -- get the code from the stack
	    CurrentIP = 0 - CurrentIP
	    FullInstr = pop_stack ()
	 else
	     -- just execute the next instruction in instruction space
	    if CurrentIP > length(InstructionSpaceTable[ThreadPointer]) or CurrentIP < 1 then
		puts(1,"\n*** ETSE Error Detector v.01 ***\n")
		puts(1,"Code: PTR01\n")
		printf(1,"An error was detected in thread #%d(%s)\n",{threadID(ThreadPointer),
								      getThreadName(threadID(ThreadPointer))})
		puts(1,"Attempt to execute instruction beyond end or beginning of code\n")
		showIP()
		terminate(ThreadPointer)
		return
	    end if
	    FullInstr =InstructionSpaceTable[ThreadPointer][CurrentIP]
	 end if

	  -- make sure the actual instruction is also valid
	 if (atom (FullInstr)) then
	    TopazError ('p', "topaz_step", "Attempt to execute an " &
			"improperly formed instruction")
	 end if

	 IPWasChanged = 0 -- FALSE
	 if length(positions[ThreadPointer])=2 then
	    positions[ThreadPointer] &= gfxMode
	 end if
	 if gfxMode != positions[ThreadPointer][3] then
	    positions[ThreadPointer]={1,1,gfxMode}
	 end if
	 if and_bits(ThreadFlags[ThreadPointer],FLAGS_DONTKEEP) != FLAGS_DONTKEEP then
	     position(positions[ThreadPointer][1],positions[ThreadPointer][2])
	 end if
	 if killed or NumThreads=0 then
	    killed=0 return
	 else
	     if errorChecking then
		check_thread(ThreadIDTable[ThreadPointer],CurrentIP)
	    end if
	    TopazExec (FullInstr)
	    positions[ThreadPointer]=get_position()&gfxMode
		executed[ThreadPointer]=1
	    if CurrentIP > 0 then pointers[ThreadPointer]=CurrentIP end if
	 end if
      end if -- end of check for suspended thread

      if (CurrThreadIsFinished) then
	  -- we hit a 'tEND' instruction; kill this thread
	  -- (TP is automatically incremented & untrapped)
	 kill_thread (get_thread_id())
      else
	  -- no 'tEND' was hit, so continue as normal
	  -- increment the IP (unless already modified)
	 if (IPWasChanged = 0) then
	    if (CurrentIP > 0) then -- otherwise, we encountered tIF
	       CurrentIP = CurrentIP + 1
	    end if
	 end if
	  -- save the new IP
	 InstructionPointerTable[ThreadPointer] = CurrentIP
	  -- increment the TP (unless trapped)
	 if (TP_IsTrapped = 0) then
	    TimeTable[ThreadPointer] += 1
	    TickCount += 1
	    IncrementThreadPointer ()
	 end if
      end if

   else

       -- the thread pointer is invalid (likely no threads), so fix
      ThreadPointer = 1

   end if -- end of check for valid thread pointer

end procedure -- end topaz_step()

 --------------------------------------------------
 -- local routines for the built-in instructions --
 --------------------------------------------------

without warning

 --
 -- tADD: Pops the top two elements of the stack, adds them together,
 --    and pushes the result back onto the stack. No parameters.
 --
procedure add (sequence Params)

   object Addend1, Addend2

   Addend1 = pop_stack ()
   Addend2 = pop_stack ()

   push_stack (Addend1 + Addend2)

end procedure -- end add()


 --
 -- tSUB: Pops the top two elements of the stack, subtracts the first
 --    from the second, and pushes the result back onto the stack. No
 --    params.
 --
procedure sub (sequence Params)

   object Num1, Num2

   Num1 = pop_stack ()
   Num2 = pop_stack ()

   push_stack (Num1 - Num2)

end procedure -- end sub()


 --
 -- tMUL: Pops the top two elements of the stack, multiplies them
 --    together, and pushes the result back onto the stack. No params.
 --
procedure mul (sequence Params)

   object Factor1, Factor2

   Factor1 = pop_stack ()
   Factor2 = pop_stack ()

   push_stack (Factor1 * Factor2)

end procedure -- end mul()


 --
 -- tDIV: Pops the top two elements of the stack, divides the first by
 --    the second, and pushes the result back onto the stack. No
 --    params.
 --
procedure div (sequence Params)

   sequence answers
   object Numerator, Denominator
   answers={}
   Numerator   = pop_stack ()
   Denominator = pop_stack ()

   push_stack (Numerator / Denominator)

end procedure -- end div()


 --
 -- tFLOOR: Pops the top element of the stack, performs a Euphorian
 --    'floor' on it, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_floor (sequence Params)

   object Obj

   Obj = pop_stack ()

   push_stack (floor (Obj))

end procedure -- end topaz_floor()


 --
 -- tCAT: Pops the top two elements of the stack, concatenates them
 --    (applies a Euphorian '&'), and pushes the result back onto the
 --    stack. No params.
 --
procedure seq_concat (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 & Obj2)

end procedure -- end seq_concat()


 --
 -- tEQ: Pops the top two elements of the stack, performs a
 --    Euphorian '=' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure equal_sign (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (equal(Obj1, Obj2))

end procedure -- end equal_sign()


 --
 -- tLESS: Pops the top two elements of the stack, performs a
 --    Euphorian '<' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure less_than (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 < Obj2)

end procedure -- end less_than()


 --
 -- tGRTR: Pops the top two elements of the stack, performs a
 --    Euphorian '>' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure greater_than (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 > Obj2)

end procedure -- end greater_than()


 --
 -- tNEQ: Pops the top two elements of the stack, performs a
 --    Euphorian '!=' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure not_equal (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 != Obj2)

end procedure -- end not_equal()


 --
 -- tAND: Pops the top two elements of the stack, applies a Euphorian
 --    'and' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_and (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 and Obj2)

end procedure -- end topaz_and()


 --
 -- tOR: Pops the top two elements of the stack, applies a Euphorian
 --    'or' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_or (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 or Obj2)

end procedure -- end topaz_or()


 --
 -- tXOR: Pops the top two elements of the stack, applies a Euphorian
 --    'xor' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_xor (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 xor Obj2)

end procedure -- end topaz_xor()


 --
 -- tNOT: Pops the top element of the stack, applies a Euphorian 'not'
 --    to it, and pushes the result back onto the stack. No params.
 --
procedure topaz_not (sequence Params)

   object Obj

   Obj = pop_stack ()

   push_stack (Obj = 0)

end procedure -- end topaz_not()


 --
 -- tCOMP: Pops the top two elements of the stack, compares them, and
 --    pushes the result (-1, 0, or 1) back onto the stack. No params.
 --
procedure topaz_compare (sequence Params)

   object Obj1, Obj2

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (compare (Obj1, Obj2))

end procedure -- end topaz_compare()


 --
 -- tPUSHA: Takes the data located at the specified virtual address,
 --    and pushes it onto the stack. If there are multiple addresses,
 --    this is repeated for each. One or more params.
 --
procedure push_vaddr (sequence Params)

   for i = 1 to length (Params) do
      push_stack (get_vaddress(Params[i]))
   end for

end procedure -- end push_vaddr()


 --
 -- tPUSHD: Takes the parameter, and pushes it onto the stack. If
 --    there is more than one, this is repeated for each. One or more
 --    params.
 --
procedure push_data (sequence Params)

   for i = 1 to length (Params) do
      push_stack (Params[i])
   end for

end procedure -- end push_data()


 --
 -- tPOP: Pops the top element from the stack and places it in the
 --    specified virtual address. If there are multiple addresses,
 --    the process is repeated for each. If a virtual address is 0,
 --    the top element of the stack is discarded. One or more params.
 --
procedure topaz_pop (sequence Params)

   object  Obj
   integer VAddr

   for i = 1 to length (Params) do
      Obj = pop_stack ()
      VAddr = Params[i]
      if (VAddr != 0) then
	 put_vaddress (Obj, VAddr)
      end if
   end for

end procedure -- end topaz_pop()


 --
 -- tCOPY: Creates a duplicate of the top stack element, and pushes
 --    it onto the stack. No params.
 --
procedure topaz_copy (sequence Params)

   object Obj

   Obj = pop_stack ()

   push_stack (Obj)
   push_stack (Obj)

end procedure -- end topaz_copy()


 --
 -- tIF: Pops the top element of the stack. Then executes its
 --    parameter (which is a single instruction, properly wrapped with
 --    its own parameters) if that element is not zero. One param.
 --
procedure if_not_0 (sequence Params)

   object Obj

   Obj = pop_stack ()

   if (atom (Obj)) then
      if (Obj != 0) then
	 push_stack (Params[1])
	 CurrentIP = 0 - CurrentIP
      end if
   end if

end procedure -- end if_not_0()


global procedure suspend(integer id)
    IsSuspendedTable[find(id,ThreadIDTable)] = 1 -- TRUE
    SuspendedBy[find(id,ThreadIDTable)]=getThreadID(ThreadPointer)
end procedure
global procedure resume(integer id)
    IsSuspendedTable[find(id,ThreadIDTable)] = 0 -- TRUE
end procedure

 --
 -- tTRAP: Traps the thread pointer, causing the Topaz engine to
 --    dedicate all execution steps to the current thread (until it
 --    untraps the pointer). No params.
 --
procedure trap_tp (sequence Params)

   integer tID
   if errorChecking=0 then
      if (TP_TrapsAllowed) then
       TP_IsTrapped = 1 --TRUE
      end if
   else
      for z = 2 to length(ThreadIDTable) do
	if ThreadIDTable[z] != ThreadPointer then
	    suspend(ThreadIDTable[z])
	end if
      end for
   end if

end procedure -- end trap_tp()


 --
 -- tUTRAP: Untraps the thread pointer, allowing other threads to
 --    run concurrently. No params.
 --
procedure untrap_tp (sequence Params)

   if errorChecking=0 then
    TP_IsTrapped = 0 -- FALSE
   else
    for z = 2 to length(ThreadIDTable) do
	if ThreadIDTable[z] != ThreadPointer then
	    resume(ThreadIDTable[z])
	end if
    end for
   end if

end procedure -- end untrap_tp()


 --
 -- tSUSP: Pops the top element of the stack, then suspends the
 --    thread with matching ID. If the current thread suspends
 --    itself, then the thread pointer is untrapped. No params.
 --
procedure suspend_thread (sequence Params)

   object  ThreadID
   integer TargetThread

   ThreadID = pop_stack ()

   TargetThread = find (ThreadID, ThreadIDTable)

   if (TargetThread = 0) then
      TopazError ('p', "suspend_thread", "Unknown thread ID given")
   end if

   IsSuspendedTable[TargetThread] = 1 -- TRUE
   SuspendedBy[TargetThread]=ThreadPointer

   if (ThreadID = get_thread_id ()) then
      TP_IsTrapped = 0 -- FALSE
   end if
   TickCount-=TimeTable[TargetThread]
end procedure -- end suspend_thread()

 --
 -- tRESUM: Pops the top element of the stack, and identifies the
 --    thread with a matching Id. Then sets the specified thread's
 --    flag so that it is no longer suspended. NOTE: A thread is
 --    incapable of resuming itself.
 --
procedure resume_thread (sequence Params)

   object  ThreadID
   integer TargetThread,totalPriority,timeAdjust
   ThreadID = pop_stack ()

   TargetThread = find (ThreadID, ThreadIDTable)

   if (TargetThread = 0) then
      TopazError ('p', "resume_thread", "Unknown thread ID given")
   end if

   totalPriority=0
   for i=1 to NumThreads do
      if IsSuspendedTable[i]=0 then  totalPriority+=PriorityTable[i] end if
   end for
   if totalPriority=0 then
      timeAdjust=0
   else
      timeAdjust=floor(PriorityTable[TargetThread]/totalPriority
      *TickCount+.5)
   end if
   IsSuspendedTable[TargetThread] = 0 -- FALSE
   TimeTable[TargetThread]=timeAdjust
   TickCount+=timeAdjust


end procedure -- end resume_thread()


 --
 -- tGETID: Pushes the thread's unique ID onto the stack. No params.
 --
procedure topaz_get_thread_id (sequence Params)

   push_stack (get_thread_id ())

end procedure -- end topaz_get_thread_id()


 --
 -- tINCIP: Increments the instruction pointer for the thread by the
 --    amount of that parameter; use negative values to jump/branch
 --    backwards. If the parameter is 0, pops the top element of the
 --    stack and uses that value instead. One param.
 --
procedure topaz_inc_ip (sequence Params)

   object Value

   if length(Params) then
    Value = Params[1]
   else
    Value=pop_stack()
   end if

   inc_ip (Value)

end procedure -- end topaz_inc_ip()


 --
 -- tDISP: Pops the top element of the stack and displays it.
 --
procedure topaz_display (sequence Params)

   ? pop_stack ()

end procedure


 --
 -- tNOP: This instruction intentionally does nothing. No params.
 --
procedure topaz_nop (sequence Params)

    -- this is meant to be a completely empty procedure

end procedure -- end topaz_nop()


 --
 -- tEND: Ends the current thread and causes it to be killed. The
 --    thread pointer is also untrapped. No params.
 --
global procedure topaz_end (sequence Params)

   CurrThreadIsFinished = 1 -- TRUE

   TP_IsTrapped = 0 -- FALSE

end procedure -- end topaz_end()
procedure prunt(integer fn,object dat)
    print(fn,dat)
    puts(fn,'\n')
end procedure
function getTo0(integer fn)
    sequence s
    integer c
    s=""
    while 1 do
	c=getc(fn)
	if c < 1 then
	    exit
	end if
	s &= c
    end while
    return s
end function
global procedure saveState(sequence file)
    sequence scrn,vc,pos
    integer fn,mode,rvisible
    fn=open(file,"wb")
    if fn=-1 then
	return
    end if
    pos=get_position()
    vc=video_config()
    mode=vc[VC_MODE]
    if mode <= 3 or mode=7 then
	scrn=save_text_image({1,1},vc[3..4])
    else
	rvisible=rat_visible
	hide_rat()
	scrn=save_image({0,0},vc[5..6])
	if rvisible then
	    show_rat()
	end if
    end if
    puts(fn,"VMS") -- VMS = Virtual Machine Status
    prunt(fn,mode)
    prunt(fn,length(scrn[1]))
    prunt(fn,length(scrn))
    for y = 1 to length(scrn) do
	puts(fn,scrn[y])
    end for
    prunt(fn,pos)
    prunt(fn,InstructionSpaceTable)
    prunt(fn,StackSizeTable)
    prunt(fn,StackTable)
    prunt(fn,DataSpaceTable)
    prunt(fn,pointers)
    prunt(fn,executed)
    prunt(fn,ThreadIDTable)
    prunt(fn,InstructionPointerTable)
    prunt(fn,InstructionRoutineTable)
    prunt(fn,IsSuspendedTable)
    prunt(fn,SuspendedBy)
    prunt(fn,PriorityTable)
    prunt(fn,TimeTable)
    prunt(fn,ThreadNames)
    prunt(fn,sIDs)
    prunt(fn,sIDs2)
    prunt(fn,keyIDs)
    prunt(fn,tmrIDs)
    prunt(fn,timers)
    prunt(fn,positions)
    print(fn,rat_button) puts(fn,'\n')
    print(fn,rat_x) puts(fn,'\n')
    print(fn,rat_y) puts(fn,'\n')
    puts(fn,current_dir()&0)
    puts(fn,length(files))
    for z = 1 to length(files) do
	puts(fn,length(files[z])&files[z]&length(modes[z])&modes[z])
	prunt(fn,fpos[z])
    end for
    prunt(fn,msgs)
    prunt(fn,vars)
    prunt(fn,vlist)
    prunt(fn,ThreadFlags)
    puts(fn,ThreadPointer&execute)
    prunt(fn,funcs)
    print(fn,MaxInstruction)
    puts(fn,'\n'&TP_IsTrapped&TP_TrapsAllowed)
    print(fn,NumThreads)
    puts(fn,'\n')
    print(fn,NextThreadID)
    puts(fn,'\n')
    print(fn,CurrentIP)
    puts(fn,'\n'&IPWasChanged&CurrThreadIsFinished)
    print(fn,TickCount)
    close(fn)
end procedure
function getSomething(integer fn,object original)
    sequence dat
    dat=get(fn)
    if dat[1] = GET_SUCCESS then
	return dat[2]
    end if
    return original
end function
function stToIntMyWay(sequence s)
    s=value(s)
    return s[2]
end function
function getInt(integer fn)
    object ln
    ln=gets(fn)
    if atom(ln) then
	return 0
    end if
    ln=stToIntMyWay(ln)
    return ln
end function
global function loadState(sequence fil)
    sequence sig,temp,scrn,pos,fils,fmode,fps,dr
    integer fn,wid,len,ps,mode,rbutton,rx,ry
    object tmp
    fps={}
    fils={}
    scrn={}
    opened={}
    modes={}
    files={}
    fpos={}
    fn=open(fil,"rb")
    if fn=-1 then
	return 1
    end if
    sig=get_bytes(fn,3)
    if not equal(sig,"VMS") then
	return 2
    end if
    mode=getInt(fn)
    wid=getInt(fn)
    len=getInt(fn)
    for y = 1 to len do
	scrn &= {get_bytes(fn,wid)}
    end for
    pos=getSomething(fn,{1,1})
    InstructionSpaceTable=getSomething(fn,InstructionSpaceTable)
    StackSizeTable=getSomething(fn,StackSizeTable)
    StackTable=getSomething(fn,StackTable)
    DataSpaceTable=getSomething(fn,DataSpaceTable)
    pointers=getSomething(fn,pointers)
    executed=getSomething(fn,executed)
    ThreadIDTable=getSomething(fn,ThreadIDTable)
    InstructionPointerTable=getSomething(fn,InstructionPointerTable)
    InstructionRoutineTable=getSomething(fn,InstructionRoutineTable)
    IsSuspendedTable=getSomething(fn,IsSuspendedTable)
    SuspendedBy=getSomething(fn,SuspendedBy)
    PriorityTable=getSomething(fn,PriorityTable)
    TimeTable=getSomething(fn,TimeTable)
    ThreadNames=getSomething(fn,ThreadNames)
    sIDs=getSomething(fn,sIDs)
    sIDs2=getSomething(fn,sIDs2)
    keyIDs=getSomething(fn,keyIDs)
    tmrIDs=getSomething(fn,tmrIDs)
    timers=getSomething(fn,timers)
    positions=getSomething(fn,positions)
    rbutton=getSomething(fn,rat_button)
    rx=getSomething(fn,rat_x)
    ry=getSomething(fn,rat_y)
    dr=getTo0(fn)
    len=getc(fn)
    for z = 1 to len do
	len=getc(fn)
	fils &= {get_bytes(fn,len)}
	len=getc(fn)
	fmode=get_bytes(fn,len)
	fps &= getInt(fn)
    end for
    --opened=getSomething(fn,opened)
    msgs=getSomething(fn,msgs)
    vars=getSomething(fn,vars)
    vlist=getSomething(fn,vlist)
    ThreadFlags=getSomething(fn,ThreadFlags)
    ThreadPointer=getSomething(fn,ThreadPointer)
    execute=getc(fn)
    funcs=getSomething(fn,funcs)
    MaxInstruction=getInt(fn)
    TP_IsTrapped=getc(fn)
    TP_TrapsAllowed=getc(fn)
    NumThreads=getInt(fn)
    NextThreadID=getInt(fn)
    CurrentIP=getInt(fn)
    IPWasChanged=getc(fn)
    CurrThreadIsFinished=getc(fn)
    TickCount=getInt(fn)
    close(fn)
    if graphics_mode(mode) then
    end if
    if mode <= 3 or mode=7 then
	display_text_image({1,1},scrn)
    else
	display_image({0,0},scrn)
	init_rat()
	rat_button=rbutton
	rat_x=rx rat_y=ry
	show_rat()
    end if
    position(pos[1],pos[2])
    if chdir(dr) then
    end if
    for z = 1 to length(fps) do
	opened &= open(fils[z],fmode)
	if seek(opened[length(opened)],fps[z]) then
	end if
    end for
    fpos=fps
    success=1
    return 0
end function
global function isSuspended(integer id)
    if find(id,ThreadIDTable) then
	return IsSuspendedTable[find(id,ThreadIDTable)]
    end if
    return 0
end function

with warning

 ----------------------
 -- initialize Topaz --
 ----------------------

 -- basic initialization

InitTopaz ()

 -- load the "built-in" instructions & assign to constants
global constant tADD   = add_instruction (routine_id ("add"))
global constant tSUB   = add_instruction (routine_id ("sub"))
global constant tMUL   = add_instruction (routine_id ("mul"))
global constant tDIV   = add_instruction (routine_id ("div"))
global constant tFLOOR = add_instruction (routine_id ("topaz_floor"))
global constant tAND   = add_instruction (routine_id ("topaz_and"))
global constant tCAT   = add_instruction (routine_id ("seq_concat"))
global constant tOR    = add_instruction (routine_id ("topaz_or"))
global constant tXOR   = add_instruction (routine_id ("topaz_xor"))
global constant tNOT   = add_instruction (routine_id ("topaz_not"))
global constant tCOMP  = add_instruction
			    (routine_id ("topaz_compare"))
global constant tEQ    = add_instruction (routine_id ("equal_sign"))
global constant tLESS  = add_instruction (routine_id ("less_than"))
global constant tGRTR  = add_instruction (routine_id ("greater_than"))
global constant tNEQ   = add_instruction (routine_id ("not_equal"))
global constant tPUSHA = add_instruction (routine_id ("push_vaddr"))
global constant tPUSHD = add_instruction (routine_id ("push_data"))
global constant tPOP   = add_instruction (routine_id ("topaz_pop"))
global constant tCOPY  = add_instruction (routine_id ("topaz_copy"))
global constant tIF    = add_instruction (routine_id ("if_not_0"))
global constant tTRAP  = add_instruction (routine_id ("trap_tp"))
global constant tUTRAP = add_instruction (routine_id ("untrap_tp"))
global constant tSUSP  = add_instruction
			    (routine_id ("suspend_thread"))
global constant tRESUM = add_instruction
			    (routine_id ("resume_thread"))
global constant tGETID = add_instruction
			    (routine_id ("topaz_get_thread_id"))
global constant tINCIP = add_instruction (routine_id ("topaz_inc_ip"))
global constant tDISP  = add_instruction
			    (routine_id ("topaz_display"))
global constant tNOP   = add_instruction (routine_id ("topaz_nop"))
global constant tEND   = add_instruction (routine_id ("topaz_end"))

 -- establish other useful global constants
global constant TOPAZ_VERSION   = 1.0 -- indicates the version number
global constant INC_FROM_STACK  = 0   -- for use with tINCIP
global constant NULL_VADDR      = 0   -- for use with tPOP
killID=routine_id("kill_thread")
include topaz2.e

